#include <iostream>
#include <stdlib.h>
#include <math.h>
// -------------------------------------------------------------------------

using namespace std;
// -------------------------------------------------------------------------

/// \brief Zwraca kolejna liczbe pierwsza mniejsza od podanej
///
/// \param refval - dodatnia wartosc referencyjna (zakladamy ze jest to liczba pierwsza)
/// \returns Zwraca kolejna liczbe pierwsza mniejsza od podanej, jezeli podana liczba jest nieprawidlowa zwraca 0
int findnextprime(int refval)
{
    if(refval < 0)                      // zawsze mozemy podac abs(refval)
        return 0;

    refval--;                           // zakladamy ze refval jest wartoscia referencyjna
    if((refval%2) == 0)                 // jezeli pazysta obliczenie wartosci
        refval--;

    for(int p=refval;p>2;p-=2)
    {
        // sprawdzamy czy p jest liczba pierwsza
        bool t = false;
        int sq = sqrt(p);
        for(int i=3;(i<=sq) && (!t);i+=2)
            t = ((p%i) == 0);

        // czy znaleziono podzielnik p
        if(!t)
            return p;                   // jezeli p nie ma podzilenika to zwracamy ja jako liczbe pierwsza
    }

    return 0;
}
// -------------------------------------------------------------------------

/// \brief Oblicza odleglosc podanej wartosci od najblizszej potegi 2
///
/// \param val - wartosc dla ktorej odleglosc chcemy obliczyc
/// \returns wartosc dodatnia okreslajaca odleglosc podanej wartosci od najblizszej potegi 2
int calculate2forcedistance(int val)
{
    int f1, f2 = 1;
    while(f2 <= val)
    {
        f1 = f2;
        f2 = f2 << 1;
    }

    if(f2 == val) // jezeli liczba jest potega dwojki
        return 0;

    int d1 = f2 - val;
    int d2 = val - f1;
    int mindist = d1<d2?d1:d2;

    return mindist;
}
// -------------------------------------------------------------------------

/// \brief Funckja oceniajaca podana liczba pierwsza
///
/// \param size - ilosc danych ktore chcemy przechowywac
/// \param avelistsize - srednia zakladana dlugosc listy
/// \param prime - dana liczba pierwsza
/// \returns wartosc oceny danej liczby pierwszej (im wieksza tym lepsza)
int assess(int size, int avelistsize, int prime)
{
    int tsize = size/avelistsize;                   // idealna ilosc kluczy
    int tsizedist = abs(tsize-prime);               // odleglosc liczby pierwszej od wartosci idealnej
    int f2dist = calculate2forcedistance(prime);    // odleglosc od najblizszej potegi 2
    return f2dist - tsizedist;                      // wartosc oceny
}
// -------------------------------------------------------------------------

/// \brief Oblicza możliwie najlepsza wartosc modulo dla haszowania modularnego
///
/// \param size - zakladana ilosc przechowywanych danych
/// \param avelistsize - zakladana srednia dlugosc listy
/// \returns mozliwie najlepsza wartosc modulo dla haszowania modularnego
int finduniversumsize(int size, int avelistsize)
{
    int prime = findnextprime(size-1);
    int besta = assess(size, avelistsize, prime);
    int bestp = prime;

    while(prime > 0)
    {
        prime = findnextprime(prime);
        int a = assess(size, avelistsize, prime);
        if(besta < a)
        {
            besta = a;
            bestp = prime;
        }
    }
    return bestp;
}
// -------------------------------------------------------------------------

int main(void)
{
    int size, avelistsize;
    cout << "Podaj ile wartosc chcesz przechowywac: ";
    cin >> size;
    cout << "Podaj srednia dlugosc listy: ";
    cin >> avelistsize;
    cout << "Rozmiar universum: " << finduniversumsize(size, avelistsize) << endl;

    return 0;
}
// -------------------------------------------------------------------------
